import os
import inspect


def load_prompt(prompt):
    file_path = os.path.abspath(inspect.getfile(inspect.currentframe()))
    dir_path = os.path.dirname(file_path)
    path = f"{dir_path}/{prompt}"
    with open(path, "r") as f:
        return f.read()


from typing import Optional,List
import pinecone,re
import os
import uuid
import numpy as np
import logging
from pydantic import Field

from langchain.vectorstores import Pinecone
from langchain.embeddings.openai import OpenAIEmbeddings
from langchain.chat_models import ChatOpenAI
from langchain.load.serializable import Serializable
from autoxx.utils.processing_text import split_text
from langchain.chains import HypotheticalDocumentEmbedder
from autoxx.tools.knowledge_base.prompts import load_prompt
from langchain.schema import HumanMessage, SystemMessage, AIMessage, BaseMessage
logger = logging.getLogger(__name__)

class Document(Serializable):
    id: str = None
    page_content: str
    metadata: dict = Field(default_factory=dict)
    score: float = None

class knowleage_application:
    def __init__(self, corpus:str, namespace:Optional[str] = None, model:str = "gpt-3.5-turbo-16k", embedding_model: str="text-embedding-ada-002"):
        if not is_valid_corpus_name(corpus):
            raise ValueError(f"Invalid corpus name: {corpus}. coprus name must consist of lower case alphanumeric characters or '-', and must start and end with an alphanumeric character")
        
        self.corpus = corpus
        self.namespace=namespace
        self._text_key = "text"
        self._kb_id_key = "_internal_kb_id"

        pinecone_api_key = os.getenv("PINECONE_API_KEY")
        assert pinecone_api_key is not None, "Please set PINECONE_API_KEY environment variable"
        pinecone_environment =  os.getenv("PINECONE_ENVIRONMENT") or "us-central1-gcp"

        pinecone.init(api_key=pinecone_api_key, environment=pinecone_environment)
        index_list = pinecone.list_indexes()
        if self.corpus not in index_list:
            logger.info(f"vector index {self.corpus} not found, creating...")
            pinecone.create_index(name=self.corpus, metric="cosine", dimension=1536)
        self.pinecone_index = pinecone.Index(self.corpus)

        self.embeddings = OpenAIEmbeddings(model=embedding_model)
        self.llm = ChatLLM(model_name=model)
        self.hyde_embeddings = HypotheticalDocumentEmbedder.from_llm(self.llm, self.embeddings, "web_search")
        self.vectorstore = Pinecone(self.pinecone_index, self.embeddings.embed_query, self._text_key, self.namespace)
        self.hyde_vectorstore = Pinecone(self.pinecone_index, self.hyde_embeddings.embed_query, self._text_key, self.namespace)


    def query(self, query:str, enable_hype:bool = True, chat_history: Optional[list[dict[str, str]]]=None):
        """
        chat_history: list of dict, each dict has two keys: "role" (ai, human, system) and "content"
        """
        docs = self.similarity_search(query, enable_hype)

        system_prompt = load_prompt("qa")
        for index, doc in enumerate(docs):
            system_prompt += f"Document_{index}: {doc.page_content}\n\n"

        messages = [
            SystemMessage(content=system_prompt),
        ]

        if chat_history is not None and len(chat_history) > 0:
            messages.extend(messages_from_dict(chat_history))

        messages.append(HumanMessage(content=query))
        
        answer = self.llm(messages).content

        return {"answer": answer, "reference": docs}

    def similarity_search(self, query:str, enable_hype:bool = True, retrieve_top_k:int = 3):
        res = []
        if enable_hype:
            docs = self.hyde_vectorstore.similarity_search_with_score(query, retrieve_top_k)
        else:
            docs = self.vectorstore.similarity_search_with_score(query, retrieve_top_k)

        for doc, score in docs:
            res.append(Document(
                page_content=doc.page_content,
                metadata=doc.metadata,
                score=score
            ))
        return res

    def upsert_document(self, docs: List[Document]) -> List[str]:
        texts = []
        medadatas = []
        ids = []
        for doc in docs:
            doc.id = doc.id or str(uuid.uuid4())
            doc.metadata[self._kb_id_key] = doc.id
            chunks = [chunk for chunk, _ in (split_text(doc.page_content))]
            if len(chunks) == 1:
                texts.append(doc.page_content)
                medadatas.append(doc.metadata)
                ids.append(doc.id)
                continue
            for chunk_index, chunk in enumerate(chunks):
                chunk_id = doc.id + f"-{chunk_index}"
                texts.append(chunk)
                chunk_metadata = {"chunk_id":chunk_id}
                chunk_metadata.update(doc.metadata)
                medadatas.append(chunk_metadata)
                ids.append(chunk_id)

        return self.vectorstore.add_texts(texts, medadatas, ids=ids, namespace=self.namespace)

    def delete_document(self, document_ids: Optional[List[str]] = None, internal_kb_ids: Optional[List[str]]=None) -> None:
        if document_ids is not None and len(document_ids) > 0:
            self.vectorstore.delete(
                ids=document_ids,
            )
        
        if internal_kb_ids is not None and len(internal_kb_ids) > 0:
            self.pinecone_index.delete(
                filter={self._kb_id_key: {"$in": internal_kb_ids}},
                namespace=self.namespace,
            )

    def retrieve_document(self, document_ids: Optional[List[str]] = None, internal_kb_ids: Optional[List[str]]=None) ->  List[Document]:
        docs = []
        # use fetch api to fecth specific documents
        if document_ids is not None and len(document_ids) > 0:
            response =  self.pinecone_index.fetch(ids=document_ids, namespace=self.namespace)
            for id, vector in response['vectors'].items():
                metadata = vector['metadata']
                if self._text_key in metadata:
                    text = metadata.pop(self._text_key)
                    docs.append(Document(id=id, page_content=text, metadata=metadata))
                else:
                    logger.warning(
                        f"Found document with no `{self._text_key}` key. Skipping."
                    )
            return docs

        filter = None
        if internal_kb_ids is not None and len(internal_kb_ids) > 0:
            filter={self._kb_id_key: {"$in": internal_kb_ids}}

        # try to load all documents in the index   
        index_description = self.pinecone_index.describe_index_stats()
        vector_count = index_description["total_vector_count"]
        if self.namespace in index_description["namespaces"]:
            vector_count = index_description["namespaces"][self.namespace]["vector_count"]

        logger.info(f"retrieving in namespace {self.namespace}, vector count: {vector_count}")
        if vector_count == 0:
            return docs
    
        response = self.pinecone_index.query(
            top_k=vector_count,
            vector=list(np.zeros(1536)),
            namespace=self.namespace,
            fikter=filter,
            include_metadata=True,
        )

        for res in response["matches"]:
            metadata = res["metadata"]
            id = res["id"]
            if self._text_key in metadata:
                text = metadata.pop(self._text_key)
                docs.append(Document(id=id, page_content=text, metadata=metadata))
            else:
                logger.warning(
                    f"Found document with no `{self._text_key}` key. Skipping."
                )
        return docs

def is_valid_corpus_name(corpus_name):
    # Regular expression pattern for the corpus name rule
    pattern = r'^[a-z0-9][a-z0-9-]*[a-z0-9]$'

    # Check if the corpus name matches the pattern
    if re.match(pattern, corpus_name):
        return True
    else:
        return False
    
def ChatLLM(
    model_name="gpt-3.5-turbo",
    temperature=0,
    request_timeout=120,
):
    if model_name.startswith("gpt-4") and os.getenv("GPT_4_MODEL_API_KEY") is not None:
        return ChatOpenAI(
            model_name=model_name,
            temperature=temperature,
            request_timeout=request_timeout,
            model_kwargs={
                "api_key": os.getenv("GPT_4_MODEL_API_KEY"),
                "api_base": os.getenv("GPT_4_MODEL_API_BASE", "https://api.openai.com/v1"),
                "api_type": os.getenv("GPT_4_MODEL_API_TYPE", "open_ai"),
                "api_version": os.getenv("GPT_4_MODEL_API_VERSION"),
                "deployment_id": os.getenv("GPT_4_MODEL_API_DEPLOYMENT_ID"),
            }
        )
    else:
        return ChatOpenAI(
            model_name=model_name,
            temperature=temperature,
            request_timeout=request_timeout,
        )

def _message_from_dict(message: dict) -> BaseMessage:
    _type = message["role"]
    if _type == "human":
        return HumanMessage(message["content"])
    elif _type == "ai":
        return AIMessage(**message["content"])
    elif _type == "system":
        return SystemMessage(**message["content"])
    else:
        raise ValueError(f"Got unexpected type: {_type}")


def messages_from_dict(messages: List[dict]) -> List[BaseMessage]:
    """Convert messages from dict.

    Args:
        messages: List of messages (dicts) to convert.

    Returns:
        List of messages (BaseMessages).
    """
    return [_message_from_dict(m) for m in messages]



from typing import Dict, List, Optional, Type
import json, logging
from pathlib import Path
from sys import platform

from bs4 import BeautifulSoup
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.chrome.service import Service as ChromeDriverService
from selenium.webdriver.chrome.webdriver import WebDriver as ChromeDriver
from selenium.webdriver.common.by import By
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.edge.service import Service as EdgeDriverService
from selenium.webdriver.edge.webdriver import WebDriver as EdgeDriver
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.firefox.service import Service as GeckoDriverService
from selenium.webdriver.firefox.webdriver import WebDriver as FirefoxDriver
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.safari.options import Options as SafariOptions
from selenium.webdriver.safari.webdriver import WebDriver as SafariDriver
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.microsoft import EdgeChromiumDriverManager as EdgeDriverManager

from autoxx.config.config import GlobalConfig
from autoxx.utils.llm import get_chat_completion
from autoxx.utils.processing_text import split_text
from autoxx.utils.token_counter import count_message_tokens
from autoxx.utils.processing_html import extract_hyperlinks, format_hyperlinks
from autoxx.utils.base import Message

BrowserOptions = ChromeOptions | EdgeOptions | FirefoxOptions | SafariOptions
BrowserSummaryModel = "gpt-3.5-turbo-16k"
MaxSummaryTokenSize = 2048
summary_response_format = {
    "relevant_content": "Extract all information in detail relevant to question. if there is no relevant information, summarize the text briefly.",
    "relevance": "Yes/No # Whether there is a relevance",
}
formatted_summary_response_format = json.dumps(summary_response_format, indent=4)


FILE_DIR = Path(__file__).parent

def create_summary_message(chunk: str, question: str) -> Message:
    """Create a message for the chat completion

    Args:
        chunk (str): The chunk of text to summarize
        question (str): The question to answer

    Returns:
        Dict[str, str]: The message to send to the chat completion
    """
    return Message(
        role="user", 
        content=(
            f'"""{chunk}""" Analyze the above text and extract all information in detail relevant to '
            f'question: "{question}" -- if there is no relevant information, summarize the text.'
            f"\nResponse Format: \n{formatted_summary_response_format} "
            f"\nEnsure the response can be parsed by Python json.loads"
        )
    )

def no_relevance(answer: str) -> bool:
    return answer.strip().replace(".", "").lower() == "no"

def scroll_to_percentage(driver: WebDriver, ratio: float) -> None:
    """Scroll to a percentage of the page

    Args:
        driver (WebDriver): The webdriver to use
        ratio (float): The percentage to scroll to

    Raises:
        ValueError: If the ratio is not between 0 and 1
    """
    if ratio < 0 or ratio > 1:
        raise ValueError("Percentage should be between 0 and 1")
    driver.execute_script(f"window.scrollTo(0, document.body.scrollHeight * {ratio});")

def scrape_links_with_selenium(driver: WebDriver, url: str) -> list[str]:
    """Scrape links from a website using selenium

    Args:
        driver (WebDriver): The webdriver to use to scrape the links

    Returns:
        List[str]: The links scraped from the website
    """
    page_source = driver.page_source
    soup = BeautifulSoup(page_source, "html.parser")

    for script in soup(["script", "style"]):
        script.extract()

    hyperlinks = extract_hyperlinks(soup, url)

    return format_hyperlinks(hyperlinks)

def scrape_text_with_selenium(url: str) -> tuple[WebDriver, str]:
    """Scrape text from a website using selenium

    Args:
        url (str): The url of the website to scrape

    Returns:
        Tuple[WebDriver, str]: The webdriver and the text scraped from the website
    """
    logging.getLogger("selenium").setLevel(logging.CRITICAL)

    options_available: dict[str, Type[BrowserOptions]] = {
        "chrome": ChromeOptions,
        "edge": EdgeOptions,
        "firefox": FirefoxOptions,
        "safari": SafariOptions,
    }

    CFG = GlobalConfig().get()
    options: BrowserOptions = options_available[CFG.selenium_web_browser]()
    options.add_argument(
        "user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.5615.49 Safari/537.36"
    )

    if CFG.selenium_web_browser == "firefox":
        if CFG.selenium_headless:
            options.headless = True
            options.add_argument("--disable-gpu")
        driver = FirefoxDriver(
            service=GeckoDriverService(GeckoDriverManager().install()), options=options
        )
    elif CFG.selenium_web_browser == "edge":
        driver = EdgeDriver(
            service=EdgeDriverService(EdgeDriverManager().install()), options=options
        )
    elif CFG.selenium_web_browser == "safari":
        # Requires a bit more setup on the users end
        # See https://developer.apple.com/documentation/webkit/testing_with_webdriver_in_safari
        driver = SafariDriver(options=options)
    else:
        if platform == "linux" or platform == "linux2":
            options.add_argument("--disable-dev-shm-usage")
            options.add_argument("--remote-debugging-port=9222")

        options.add_argument("--no-sandbox")
        if CFG.selenium_headless:
            options.add_argument("--headless=new")
            options.add_argument("--disable-gpu")

        chromium_driver_path = Path("/usr/bin/chromedriver")

        driver = ChromeDriver(
            service=ChromeDriverService(str(chromium_driver_path))
            if chromium_driver_path.exists()
            else ChromeDriverService(ChromeDriverManager().install()),
            options=options,
        )
    driver.get(url)

    WebDriverWait(driver, 10).until(
        EC.presence_of_element_located((By.TAG_NAME, "body"))
    )

    # Get the HTML content directly from the browser's DOM
    page_source = driver.execute_script("return document.body.outerHTML;")
    soup = BeautifulSoup(page_source, "html.parser")

    for script in soup(["script", "style"]):
        script.extract()

    text = soup.get_text()
    lines = (line.strip() for line in text.splitlines())
    chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
    text = "\n".join(chunk for chunk in chunks if chunk)
    return driver, text

def summarize_chunks(text:str, question: str, url: str, driver: Optional[WebDriver] = None) -> List[str]:
    """Summarize a list of chunks

    Args:
        chunks (list[str]): The chunks to summarize
        question (str): The question to answer

    Returns:
        str: The summary of the chunks
    """
    logging.debug(f"Memorizing text:\n{'-'*32}\n{text}\n{'-'*32}\n")
    chunks = [
        chunk
        for chunk, _ in (
            split_text(text, for_model=BrowserSummaryModel)
        )
    ]

    summaries = []
    scroll_ratio = 1 / len(chunks)
    for i, chunk in enumerate(chunks):
        if driver:
            scroll_to_percentage(driver, scroll_ratio * i)

        messages = [create_summary_message(chunk, question)]
        tokens_for_chunk = count_message_tokens(messages)

        logging.info(
            f"Summarizing {url} chunk {i + 1} / {len(chunks)} of length {len(chunk)} characters, or {tokens_for_chunk} tokens"
        )

        summary_response_str = get_chat_completion(
            model=BrowserSummaryModel,
            messages=messages,
        )

        summary_response = json.loads(summary_response_str)
        if no_relevance(summary_response.get("relevance", "")):
            logging.info(f"Skipping chunk {i + 1} / {len(chunks)} of length {len(chunk)} characters, or {tokens_for_chunk} tokens, not relevant.")
            continue
        summary = summary_response.get("relevant_content", summary_response_str)

        logging.info(f"Chunk {i + 1} / {len(chunks)} summary.\n{summary}")

        summaries.append(summary)

    return summaries

def add_header(driver: WebDriver) -> None:
    """Add a header to the website

    Args:
        driver (WebDriver): The webdriver to use to add the header

    Returns:
        None
    """
    try:
        with open(f"{FILE_DIR}/js/overlay.js", "r") as overlay_file:
            overlay_script = overlay_file.read()
        driver.execute_script(overlay_script)
    except Exception as e:
        print(f"Error executing overlay.js: {e}")

def close_browser(driver: WebDriver) -> None:
    """Close the browser

    Args:
        driver (WebDriver): The webdriver to close

    Returns:
        None
    """
    driver.quit()


def browse_website(url: str, question: str) -> str | None:
    """Browse a website and return the answer and links to the user

    Args:
        url (str): The url of the website to browse
        question (str): The question asked by the user

    Returns:
        Tuple[str, WebDriver]: The answer and links to the user and the webdriver
    """
    try:
        driver, text = scrape_text_with_selenium(url=url)
    except WebDriverException as e:
        # These errors are often quite long and include lots of context.
        # Just grab the first line.
        msg = e.msg.split("\n")[0]
        return f"Error: {msg}"
    add_header(driver)

    summaries = summarize_chunks(text=text, question=question, url=url, driver=driver)
    text_gen = "\n".join(summaries)
    while count_message_tokens([create_summary_message(text_gen, question)]) > MaxSummaryTokenSize:
       summaries = summarize_chunks(text=text_gen, question=question, url=url, driver=driver)
       text_gen = "\n".join(summaries)

    if len(text_gen) == 0:
        close_browser(driver)
        return None

    links = scrape_links_with_selenium(driver, url)
    # Limit links to 5
    if len(links) > 5:
        links = links[:5]
    close_browser(driver)
    return f"Answer gathered from website: {text_gen}\n\nLinks: {links}"

def web_search(question:str, search_num:Optional[int]=3) -> List[Dict]:
    search_result = google_official_search(query=question, num_results=search_num+3)
    search_summaries = []

    index = 0
    for res in search_result:
        print(f"{res['title']} x {res['url']}\n")
        try:
            summary = browse_website(url=res['url'], question=question)
        except Exception as e:
            print(f"Error browsing websit {res['title']} x {res['url']}: {e}")
            continue

        if summary is None:
            continue

        search_summaries.append({
            "relevant_content": summary,
            "title": res['title'],
            "url": res['url']
        })

        print(f"{res['title']} x {res['url']} summary: {summary}\n")
        index += 1
        if index >= search_num:
            break

    return search_summaries

def google_official_search(query: str, num_results: int = 8) -> List[Dict]:
    """Return the results of a Google search using the official Google API

    Args:
        query (str): The search query.
        num_results (int): The number of results to return.

    Returns:
        str: The results of the search.
    """

    from googleapiclient.discovery import build
    from googleapiclient.errors import HttpError

    CFG = GlobalConfig().get()
    search_summaries = []
    try:
        # Get the Google API key and Custom Search Engine ID from the config file
        api_key = CFG.google_api_key
        custom_search_engine_id = CFG.custom_search_engine_id

        # Initialize the Custom Search API service
        service = build("customsearch", "v1", developerKey=api_key)

        # Send the search query and retrieve the results
        result = (
            service.cse()
            .list(q=query, cx=custom_search_engine_id, num=num_results)
            .execute()
        )

        # Extract the search result items from the response
        search_results = result.get("items", [])
    except HttpError as e:
        # Handle errors in the API call
        error_details = json.loads(e.content.decode())

        # Check if the error is related to an invalid or missing API key
        if error_details.get("error", {}).get(
            "code"
        ) == 403 and "invalid API key" in error_details.get("error", {}).get(
            "message", ""
        ):
            raise "Error: The provided Google API key is invalid or missing."
        else:
            raise f"Error: {e}"

    # Return the list of search result URLs
    for search_result in search_results:
        search_summaries.append({
            "relevant_content": search_result['snippet'],
            "title": search_result['title'],
            "url": search_result['link']
        })
    return search_summaries



import os
import logging
import abc
from colorama import Fore
from typing import Optional
from dataclasses import dataclass
    
@dataclass
class OpenAIConfig:
    model: str
    api_key: str
    temperature: float = 0
    api_deployment_id: Optional[str] = None
    api_type: str = "open_ai"
    api_base: str = "https://api.openai.com/v1"
    api_version: Optional[str] = None

@dataclass
class GPTLLMModelConfig(OpenAIConfig):
    temperature: float = 0

@dataclass
class GPTEmbeddingModelConfig(OpenAIConfig):
    embedding_tokenizer: str = "cl100k_base"

def check_openai_api_key(api_key: str) -> None:
    """Check if the OpenAI API key is set in config.py or as an environment variable."""
    if not api_key or api_key == "":
        logging.error(
            Fore.RED
            + "Please set your OpenAI API key in .env or as an environment variable."
        )
        exit(1)

class Config():
    """
    Configuration class to store the state of bools for different scripts access.
    """

    def __init__(self) -> None:
        """Initialize the Config class"""
        self.debug_mode = False

        self.openai_api_key = os.getenv("OPENAI_API_KEY")
        self.llm_models =   os.getenv("LLM_MODELS", "gpt-3.5-turbo,gpt-4,gpt-3.5-turbo-16k").split(",")
        self.embedding_models = os.getenv("EMBEDDING_MODELS", "text-embedding-ada-002").split(",")
        self.model_config = {}
        for model in self.llm_models:
            self.set_llm_model_config(model)
        for model in self.embedding_models:
            self.set_embedding_model_config(model)

        self.browse_spacy_language_model = os.getenv(
            "BROWSE_SPACY_LANGUAGE_MODEL", "en_core_web_sm"
        )

        self.google_api_key = os.getenv("GOOGLE_API_KEY")
        self.custom_search_engine_id = os.getenv("CUSTOM_SEARCH_ENGINE_ID")

        self.pinecone_api_key = os.getenv("PINECONE_API_KEY")
        self.pinecone_region = os.getenv("PINECONE_ENVIRONMENT")

        self.mongodb_host = os.getenv("MONGODB_HOST")
        self.mongodb_password = os.getenv("MONGODB_PASSWORD")
        self.mongodb_user = os.getenv("MONGODB_USER")

        # Selenium browser settings
        self.selenium_web_browser = os.getenv("USE_WEB_BROWSER", "chrome")
        self.selenium_headless = os.getenv("HEADLESS_BROWSER", "True") == "True"

        # User agent header to use when making HTTP requests
        # Some websites might just completely deny request with an error code if
        # no user agent was found.
        self.user_agent = os.getenv(
            "USER_AGENT",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/537.36"
            " (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
        )
    
    def set_llm_model_config(self, model_name: str):
        model_uppercase = model_name.upper()

        model = os.getenv(f"{model_uppercase}_MODEL", "gpt-3.5-turbo")
        api_key = os.getenv(f"{model_uppercase}_MODEL_API_KEY", self.openai_api_key)
        api_type = os.getenv(f"{model_uppercase}_MODEL_API_TYPE", "open_ai")
        api_base = os.getenv(f"{model_uppercase}_MODEL_API_BASE", "https://api.openai.com/v1")
        api_version = os.getenv(f"{model_uppercase}_MODEL_API_VERSION", None)
        api_deployment_id = os.getenv(f"{model_uppercase}_MODEL_API_DEPLOYMENT_ID", None)
        temperature = float(os.getenv(f"{model_uppercase}_MODEL_TEMPERATURE", 0))
        
        self.model_config[model_name] = GPTLLMModelConfig(
            model=model,
            api_key=api_key,
            api_type=api_type,
            api_base=api_base,
            api_version=api_version,
            api_deployment_id=api_deployment_id,
            temperature=temperature,
        )
        check_openai_api_key(api_type)

    def set_embedding_model_config(self, model_name:str):
        model_uppercase = model_name.upper()
        model = os.getenv(f"{model_uppercase}_MODEL",  model_name)
        api_key = os.getenv(f"{model_uppercase}_MODEL_API_KEY", self.openai_api_key)
        api_type = os.getenv(f"{model_uppercase}_MODEL_API_TYPE", "open_ai")
        api_base = os.getenv(f"{model_uppercase}_MODEL_API_BASE", "https://api.openai.com/v1")
        api_version = os.getenv(f"{model_uppercase}_MODEL_API_VERSION", None)
        api_deployment_id = os.getenv(f"{model_uppercase}_MODEL_API_DEPLOYMENT_ID", None)
        self.embedding_tokenizer = os.getenv(f"{model_uppercase}_EMBEDDING_TOKENIZER", "cl100k_base")
        
        self.model_config[model_name] = GPTEmbeddingModelConfig(
            model=model,
            api_key=api_key,
            api_type=api_type,
            api_base=api_base,
            api_version=api_version,
            api_deployment_id=api_deployment_id,
            embedding_tokenizer=self.embedding_tokenizer,
        )
        check_openai_api_key(api_key)

    def get_llm_model_config(self, model_name: str) -> GPTLLMModelConfig:
        if model_name not in self.model_config:
            raise ValueError(f"LLM Model {model_name} not found in config")
        return self.model_config[model_name]
    
    def get_embedding_model_config(self, model_name: str) -> GPTEmbeddingModelConfig:
        if model_name not in self.model_config:
            raise ValueError(f"Embedding Model {model_name} not found in config")
        return self.model_config[model_name]
    

class Singleton(abc.ABCMeta, type):
    """
    Singleton metaclass for ensuring only one instance of a class.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Call method for the singleton metaclass."""
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]
    
class GlobalConfig(metaclass=Singleton):
    def __init__(self, cfg: Optional[Config] = None) -> None:
        if cfg is None:
            cfg = Config()
        self.config = cfg

    def set(self, cfg: Config) -> None:
        self.config = cfg

    def get(self) -> Config:
        return self.config


import openai,os
from autoxx.config.config import Config, GlobalConfig

def setup_config(
        debug:bool = False,
) -> Config:
    globalCFG = GlobalConfig()

    CFG = globalCFG.get()
    openai.api_key = CFG.openai_api_key
    return CFG



import json
import traceback
import time
import threading
from typing import Dict, List, Dict, Optional
import logging

from autoxx.tools.llm.base import get_chat_completion

from autoxx.tools.web_search.search import web_search
from autoxx.tools.llm.base import llm_uils
from autoxx.utils.llm import create_message


class AGIExecutor:
    def __init__(self, objective:str, max_tasks_count: Optional[int] = 5, cancel_event: Optional[threading.Event] = None):
        if objective is None or len(objective) <= 1:  
            raise ValueError("Must provide an objective")  

        self.objective = objective
        self.max_tasks_count = max_tasks_count

        self.task_list = self.init_task_list(objective)
        self.status = "pending"
        self.error_message = None
        self.final_answer = None
        self.cancel_event = cancel_event

    def get_task_by_id(self, task_id: int):
        for task in self.task_list:
            if task["id"] == task_id:
                return task
        return None

    def init_task_list(self, objective: str) -> List[Dict]:
        task_id_counter = 0
        prompt = (
            f"You are a task management AI tasked with creating tasks for the following objective: {objective}.\n"
            f"Create new tasks based on the following plan delimited by triple backtick if necessary for the objective. Limit tasks types to those that can be completed with the available tools listed below. Task description should be detailed. "
            f"Your task: decelop a task list .\n"
            f"The current tool option is [web-search, text-completion] only. What each tool does is as follows:\n"
            f"web-search: It supports searching for information from the Internet. Interaction with web pages is not supported. For tasks using [web-search], provide the search query, and only the search query to use (eg. not 'research waterproof shoes, but 'waterproof shoes').\n"
            f"text-completion: It can be used for text extraction, summarization, copywrite, translation, etc., and it can also be used for LLM-based QA.\n"
            f"Make sure all task IDs are in chronological order.\n"
            f"For efficiency, let's think step by step and create the tasks (up to 5 tasks) that are most critical to objective. "
            f"The last step is always to provide a final formal knowledge base article (at least contains title, summary and solutions) about objective including summary of knowledge acquired.\n"
            f"Do not create any summarizing steps outside of the last step.\n"
            f"The task list output format is as follows: "
            "[{\"id\": 1, \"task\": \"Untapped Capital\", \"tool\": \"web-search\", \"dependent_task_ids\": [], \"status\": \"incomplete\", \"result\": null, \"result_summary\": null}, {\"id\": 2, \"task\": \"Consider additional insights that can be reasoned from the results of...\", \"tool\": \"text-completion\", \"dependent_task_ids\": [1], \"status\": \"incomplete\", \"result\": null, \"result_summary\": null}].\n"
            f"Ensure the output can be parsed by Python json.loads.\n task list="
        )

        messages = create_message(prompt)
        response = get_chat_completion(
            messages=messages,
            model="gpt-4"
        )
        try:
            new_tasks = json.loads(response)
        except Exception as e:
            raise ValueError(f"fail to initial task due to json decocde failed {e}. response={response}")
        task_list = []
        for new_task in new_tasks:
            task_id_counter += 1
            new_task.update({"id": task_id_counter})
            task_list.append(new_task)
        return task_list

    ### Agent functions ##############################
    def execute_task(self, task: Dict) -> Dict:
        # Check if dependent_task_id is completed
        dependent_task_ids = []
        dependent_task_prompt = ""
        if task["dependent_task_ids"]:
            if isinstance(task["dependent_task_ids"], list):
                dependent_task_ids = task["dependent_task_ids"]
            else:
                dependent_task_ids = [task["dependent_task_ids"]]

            for dependent_task_id in dependent_task_ids:
                dependent_task = self.get_task_by_id(dependent_task_id)
                if dependent_task and dependent_task["status"] != "completed":
                    return
                elif dependent_task:
                    dependent_task_prompt += f"\ntask ({dependent_task['id']}. {dependent_task['task']}) result: {dependent_task['result']}"

        # Execute task
        
        task_prompt = (
            f"Complete your assigned task based on the objective: {self.objective}. Your task: {task['task']}. "
            f"Please keep the link reference like newbing in the response.\n"
        )
        if dependent_task_prompt != "":
            task_prompt += f"The previous tasks: {dependent_task_prompt}"

        task_prompt += "\nResponse:"
        if task["tool"] == "text-completion":
            result = llm_uils("gpt-4").text_completion(task_prompt)
            summary_result_prompt = f"Please summarize the following text and and keep the reference links:\n{result}\nSummary:"
            task["result_summary"] = llm_uils().text_completion(summary_result_prompt)
        elif task["tool"] == "web-search":
            result = str(web_search(str(task['task'])))
            summary_result_prompt = f"Please summarize the following text and keep all links:\n{result}\nSummary:"
            task["result_summary"] = llm_uils().text_completion(summary_result_prompt)
        else:
            result = "Unknown tool"
        
        # Update task status and result
        task["status"] = "completed"
        task["result"] = result
        return task
    
    def execute(self):
        self.status = "running"
        # Main loop
        while any(task["status"] == "incomplete" for task in self.task_list):
            if self.cancel_event is not None and self.cancel_event.is_set():
                logging.info(f"\033[91m\033[1m CANCELLED({self.objective}) \033[0m\033[0m")
                self.status = "cancelled"
                return
              # Filter out incomplete tasks
            incomplete_tasks = [task for task in self.task_list if task["status"] == "incomplete"]
            logging.info(f"\033[95m\033[1m INCOMPLETE TASK LIST({self.objective}) \033[0m\033[0m {[str(t['id'])+': '+t['task'] + '[' + t['tool'] + ']' for t in incomplete_tasks]}")
            if incomplete_tasks:
                # Sort tasks by ID
                incomplete_tasks.sort(key=lambda x: x["id"])

                # Pull the first task
                task = incomplete_tasks[0]
                logging.info(f"\033[92m\033[1m NEXT TASK({self.objective}) \033[0m\033[0m {task['id']}:{task['task']}[{task['tool']}], depends on {task['dependent_task_ids']}")

                # Execute task & call task manager from function
                try:
                    completed_task = self.execute_task(task)
                    logging.info(f"\033[93m\033[1m TASK RESULT({self.objective}) \033[0m\033[0m {completed_task['result']}")
                except Exception as e:  
                    stack_str = "".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__))
                    logging.error(f"An error occurred while executing the task({self.objective}): {stack_str}")
                    self.status = "error"
                    self.error_message = str(e)  
                    return

            time.sleep(1)  # Sleep before checking the task list again
        self.status = "completed"

    def stat(self) -> Dict:
        return {"task_list": self.task_list, "status": self.status, "error_message": self.error_message}


"""Text processing functions"""
from math import ceil
from typing import Optional

import spacy
import tiktoken

from autoxx.config.config import GlobalConfig
from autoxx.utils.token_counter import count_string_tokens
from autoxx.utils.openai_models import OPEN_AI_MODELS

def batch(iterable, max_batch_length: int, overlap: int = 0):
    """Batch data from iterable into slices of length N. The last batch may be shorter."""
    # batched('ABCDEFG', 3) --> ABC DEF G
    if max_batch_length < 1:
        raise ValueError("n must be at least one")
    for i in range(0, len(iterable), max_batch_length - overlap):
        yield iterable[i : i + max_batch_length]

def _max_chunk_length(model: str, max: Optional[int] = None) -> int:
    model_max_input_tokens = OPEN_AI_MODELS[model].max_tokens - 1024
    if max is not None and max > 0:
        return min(max, model_max_input_tokens)
    return model_max_input_tokens


def must_chunk_content(
    text: str, for_model: str, max_chunk_length: Optional[int] = None
) -> bool:
    return count_string_tokens(text, for_model) > _max_chunk_length(
        for_model, max_chunk_length
    )


def chunk_content(
    content: str,
    for_model: str,
    max_chunk_length: Optional[int] = None,
    with_overlap=True,
):
    """Split content into chunks of approximately equal token length."""

    MAX_OVERLAP = 200  # limit overlap to save tokens

    if not must_chunk_content(content, for_model, max_chunk_length):
        yield content, count_string_tokens(content, for_model)
        return

    max_chunk_length = max_chunk_length or _max_chunk_length(for_model)

    tokenizer = tiktoken.encoding_for_model(for_model)

    tokenized_text = tokenizer.encode(content)
    total_length = len(tokenized_text)
    n_chunks = ceil(total_length / max_chunk_length)

    chunk_length = ceil(total_length / n_chunks)
    overlap = min(max_chunk_length - chunk_length, MAX_OVERLAP) if with_overlap else 0

    for token_batch in batch(tokenized_text, chunk_length + overlap, overlap):
        yield tokenizer.decode(token_batch), len(token_batch)

def split_text(
    text: str,
    for_model: str = "gpt-3.5-turbo",
    with_overlap=True,
    max_chunk_length: Optional[int] = None,
):
    """Split text into chunks of sentences, with each chunk not exceeding the maximum length

    Args:
        text (str): The text to split
        for_model (str): The model to chunk for; determines tokenizer and constraints
        max_length (int, optional): The maximum length of each chunk

    Yields:
        str: The next chunk of text

    Raises:
        ValueError: when a sentence is longer than the maximum length
    """
    max_length = _max_chunk_length(for_model, max_chunk_length)

    # flatten paragraphs to improve performance
    text = text.replace("\n", " ")
    text_length = count_string_tokens(text, for_model)

    if text_length < max_length:
        yield text, text_length
        return

    n_chunks = ceil(text_length / max_length)
    target_chunk_length = ceil(text_length / n_chunks)

    nlp: spacy.language.Language = spacy.load(GlobalConfig().get().browse_spacy_language_model)
    nlp.add_pipe("sentencizer")
    doc = nlp(text)
    sentences = [sentence.text.strip() for sentence in doc.sents]

    current_chunk: list[str] = []
    current_chunk_length = 0
    last_sentence = None
    last_sentence_length = 0

    i = 0
    while i < len(sentences):
        sentence = sentences[i]
        sentence_length = count_string_tokens(sentence, for_model)
        expected_chunk_length = current_chunk_length + 1 + sentence_length

        if (
            expected_chunk_length < max_length
            # try to create chunks of approximately equal size
            and expected_chunk_length - (sentence_length / 2) < target_chunk_length
        ):
            current_chunk.append(sentence)
            current_chunk_length = expected_chunk_length

        elif sentence_length < max_length:
            if last_sentence:
                yield " ".join(current_chunk), current_chunk_length
                current_chunk = []
                current_chunk_length = 0

                if with_overlap:
                    overlap_max_length = max_length - sentence_length - 1
                    if last_sentence_length < overlap_max_length:
                        current_chunk += [last_sentence]
                        current_chunk_length += last_sentence_length + 1
                    elif overlap_max_length > 5:
                        # add as much from the end of the last sentence as fits
                        current_chunk += [
                            list(
                                chunk_content(
                                    last_sentence,
                                    for_model,
                                    overlap_max_length,
                                )
                            ).pop()[0],
                        ]
                        current_chunk_length += overlap_max_length + 1

            current_chunk += [sentence]
            current_chunk_length += sentence_length

        else:  # sentence longer than maximum length -> chop up and try again
            sentences[i : i + 1] = [
                chunk
                for chunk, _ in chunk_content(sentence, for_model, target_chunk_length)
            ]
            continue

        i += 1
        last_sentence = sentence
        last_sentence_length = sentence_length

    if current_chunk:
        yield " ".join(current_chunk), current_chunk_length


from dataclasses import dataclass
from typing import Literal, TypedDict

MessageRole = Literal["system", "user", "assistant"]

class MessageDict(TypedDict):
    role: MessageRole
    content: str

@dataclass
class Message:
    """OpenAI Message object containing a role and the message content"""

    role: MessageRole
    content: str

    def raw(self) -> MessageDict:
        return {"role": self.role, "content": self.content}

@dataclass
class ModelInfo:
    """Struct for model information.

    Would be lovely to eventually get this directly from APIs, but needs to be scraped from
    websites for now.

    """

    name: str
    prompt_token_cost: float
    completion_token_cost: float
    max_tokens: int


@dataclass
class ChatModelInfo(ModelInfo):
    """Struct for chat model information."""

    pass


@dataclass
class TextModelInfo(ModelInfo):
    """Struct for text completion model information."""

    pass


@dataclass
class EmbeddingModelInfo(ModelInfo):
    """Struct for embedding model information."""

    embedding_dimensions: int


from autoxx.utils.llm import get_chat_completion, create_message

class llm_uils:

    model: str

    def __init__(self, model: str = "gpt-3.5-turbo"):
        self.model = model

    def text_completion(self, prompt: str) -> str:
        messages = create_message(prompt)
        response = get_chat_completion(
            messages=messages,
            model=self.model
        )

        return response



import re, time
from typing import List, Optional, Union, Sequence, Callable, Any, Dict, Tuple

from langchain.agents import Tool, LLMSingleActionAgent, AgentOutputParser, BaseSingleActionAgent, BaseMultiActionAgent
from langchain.agents.agent import ExceptionTool
from langchain.prompts import StringPromptTemplate
from langchain import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.schema import AgentAction, AgentFinish, OutputParserException
from langchain.tools.base import BaseTool
from langchain.agents.tools import InvalidTool

from autoxx.config.config import GlobalConfig
from autoxx.tools.llm.base import llm_uils


# Set up the base template
template = """Answer the following questions as best you can. You have access to the following tools:

{tools}

Use the following format:

Question: the input question you must answer
Thought: you should always think about what to do
Action: the action to take. Choose the most appropriate action from the following tool name list: [{tool_names}]. Don't take the useless tools.
Action Input: the input to the action
Observation: the result of the action
... (this Thought/Action/Action Input/Observation can repeat N times)
Thought: I now know the final answer
Final Answer: the final answer to the original input question

Begin! Give your final answer and follow the above format.

Question: {input}
{agent_scratchpad}"""

# Set up a prompt template
class CustomPromptTemplate(StringPromptTemplate):
    # The template to use
    template: str
    # The list of tools available
    tools: List[Tool]
    
    def format(self, **kwargs) -> str:
        # Get the intermediate steps (AgentAction, Observation tuples)
        # Format them in a particular way
        intermediate_steps = kwargs.pop("intermediate_steps")
        thoughts = ""
        for action, observation in intermediate_steps:
            thoughts += action.log
            thoughts += f"\nObservation: {observation}\n"
        # Set the agent_scratchpad variable to that value
        kwargs["agent_scratchpad"] = thoughts
        # Create a tools variable from the list of tools provided
        kwargs["tools"] = "\n".join([f"{tool.name}: {tool.description}" for tool in self.tools])
        # Create a list of tool names for the tools provided
        kwargs["tool_names"] = ", ".join([tool.name for tool in self.tools])
        return self.template.format(**kwargs)


class CustomOutputParser(AgentOutputParser):
    
    def parse(self, llm_output: str) -> Union[AgentAction, AgentFinish]:
        # Check if agent should finish
        if "Final Answer:" in llm_output:
            return AgentFinish(
                # Return values is generally always a dictionary with a single `output` key
                # It is not recommended to try anything else at the moment :)
                return_values={"output": llm_output.split("Final Answer:")[-1].strip()},
                log=llm_output,
            )
        # Parse out the action and action input
        regex = r"Action: (.*?)[\n]*Action Input:[\s]*(.*)"
        match = re.search(regex, llm_output, re.DOTALL|re.IGNORECASE)
        if not match:
            print(f"Could not parse LLM output: `{llm_output}`")
            return AgentAction(tool=llm_output.strip(" ").strip('"'), tool_input=llm_output.strip(" ").strip('"'), log=llm_output)
        action = match.group(1).strip()
        action_input = match.group(2)
        # Return the action and action input
        return AgentAction(tool=action, tool_input=action_input.strip(" ").strip('"'), log=llm_output)

    
def setup_planner(tools: List[Tool], model: str = "gpt-3.5-turbo") -> BaseSingleActionAgent:
    tool_names = [tool.name for tool in tools]
    output_parser = CustomOutputParser()
    config = GlobalConfig().get()

    prompt = CustomPromptTemplate(
        template=template,
        tools=tools,
        # This omits the `agent_scratchpad`, `tools`, and `tool_names` variables because those are generated dynamically
        # This includes the `intermediate_steps` variable because that is needed
        input_variables=["input", "intermediate_steps"]
    )

    model_config = config.get_llm_model_config(model)
    llm = ChatOpenAI(temperature=0, model_name=model_config.model, model_kwargs={
        "api_key": model_config.api_key,
        "api_base": model_config.api_base,
        "api_type": model_config.api_type,
        "api_version": model_config.api_version,
        "deployment_id": model_config.api_deployment_id,
    })

    # LLM chain consisting of the LLM and a prompt
    llm_chain = LLMChain(llm=llm, prompt=prompt)
    return LLMSingleActionAgent(
        llm_chain=llm_chain, 
        output_parser=output_parser,
        stop=["\nObservation:"], 
        allowed_tools=tool_names
    )

class AgentExecutor:
    """Consists of an agent using tools."""

    agent: Union[BaseSingleActionAgent, BaseMultiActionAgent]
    tools: Sequence[BaseTool]
    return_intermediate_steps: bool = False
    max_iterations: Optional[int] = 5
    max_execution_time: Optional[float] = None
    early_stopping_method: str = "force"
    handle_parsing_errors: Union[
        bool, str, Callable[[OutputParserException], str]
    ] = False
    name_to_tool_map: dict[str, BaseTool]

    def __init__(self, tools: List[Tool], model: str = "gpt-3.5-turbo"):
        self.tools = tools
        self.agent = setup_planner(tools, model=model)
        self.name_to_tool_map = {tool.name: tool for tool in self.tools}

    def _should_continue(self, iterations: int, time_elapsed: float) -> bool:
        if self.max_iterations is not None and iterations >= self.max_iterations:
            return False
        if (
            self.max_execution_time is not None
            and time_elapsed >= self.max_execution_time
        ):
            return False

        return True

    def _return(
        self,
        output: AgentFinish,
        intermediate_steps: list,
    ) -> Dict[str, Any]:
        final_output = output.return_values
        if self.return_intermediate_steps:
            final_output["intermediate_steps"] = intermediate_steps
        return final_output

    def _decide_next_step(self, inputs: Dict[str, str], intermediate_steps: List[Tuple[AgentAction, str]]) -> Union[List[AgentAction], AgentFinish]:
        try:
            # Call the LLM to see what to do.
            output = self.agent.plan(
                intermediate_steps,
                **inputs,
            )
            return output
        except OutputParserException as e:
            if isinstance(self.handle_parsing_errors, bool):
                raise_error = not self.handle_parsing_errors
            else:
                raise_error = False
            if raise_error:
                raise e
            text = str(e)
            if isinstance(self.handle_parsing_errors, bool):
                observation = "Invalid or incomplete response"
            elif isinstance(self.handle_parsing_errors, str):
                observation = self.handle_parsing_errors
            elif callable(self.handle_parsing_errors):
                observation = self.handle_parsing_errors(e)
            else:
                raise ValueError("Got unexpected type of `handle_parsing_errors`")
            output = AgentAction(ExceptionTool.name, observation, text)
            return [output]

    def _take_next_step(
        self,
        actions: List[AgentAction],
    ) -> List[Tuple[AgentAction, str]]:
        """Take a single step in the thought-action-observation loop.

        Override this to take control of how the agent makes and acts on choices.
        """
        result = []
        for agent_action in actions:
            # Otherwise we lookup the tool
            if agent_action.tool in self.name_to_tool_map:
                tool = self.name_to_tool_map[agent_action.tool]
                # We then call the tool on the tool input to get an observation
                observation = str(tool.run(agent_action.tool_input))
            else:
                observation = InvalidTool().run(agent_action.tool)
            result.append((agent_action, observation))
        return result

    def _return_stopped_response(
        self,
        question: str,
        intermediate_steps: List[Tuple[AgentAction, str]],
    ) -> AgentFinish:
        """Return response when agent has been stopped due to max iterations."""
        # `force` just returns a constant string

        thoughts = ""
        for action, observation in intermediate_steps:
            thoughts += action.log
            thoughts += f"\nObservation: {observation}\n"

        task_prompt = (
            f"Based on these factual information:\n{thoughts}\n"
            f"\nAnswer the question: {question}. Answer="
        )

        response = llm_uils("gpt-4").text_completion(task_prompt)
        return AgentFinish(
            {"output": response}, task_prompt
        )

    def run(
        self,
        input: str,
    ) -> Dict[str, Any]:
        """Run text through and get agent response."""
        inputs = {"input": input}

        intermediate_steps: List[Tuple[AgentAction, str]] = []
        # Let's start tracking the number of iterations and time elapsed
        iterations = 0
        time_elapsed = 0.0
        start_time = time.time()
        # We now enter the agent loop (until it returns something).
        while self._should_continue(iterations, time_elapsed):
            next_step = self._decide_next_step(
                inputs,
                intermediate_steps,
            )
            if isinstance(next_step, AgentFinish):
                return self._return(
                    next_step, intermediate_steps
                )
            elif isinstance(next_step, AgentAction):
                next_step = [next_step]
            
            for step in next_step:
                print(f"\nNext action:\n{step.log}\n")
            
            next_step_output = self._take_next_step(next_step)
            intermediate_steps.extend(next_step_output)
            for output in next_step_output:
                print(f"\nAction: {output[0].tool}({output[0].tool_input})\nOutput:{output[1]}\n")

            iterations += 1
            time_elapsed = time.time() - start_time
        output = self.agent.return_stopped_response(
            self.early_stopping_method, intermediate_steps, **inputs
        )
        return self._return(output, intermediate_steps)



import tiktoken
import logging
from typing import List
from autoxx.utils.base import Message

def count_message_tokens(
    messages: List[Message], model: str = "gpt-3.5-turbo"
) -> int:
    """
    Returns the number of tokens used by a list of messages.

    Args:
        messages (list): A list of messages, each of which is a dictionary
            containing the role and content of the message.
        model (str): The name of the model to use for tokenization.
            Defaults to "gpt-3.5-turbo-0301".

    Returns:
        int: The number of tokens used by the list of messages.
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        logging.warn("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if model.startswith("gpt-3.5-turbo-"):
        # !Note: gpt-3.5-turbo may change over time.
        # Returning num tokens assuming gpt-3.5-turbo-0301.")
        return count_message_tokens(messages, model="gpt-3.5-turbo")
    elif model.startswith("gpt-4-"):
        # !Note: gpt-4 may change over time. Returning num tokens assuming gpt-4-0314.")
        return count_message_tokens(messages, model="gpt-4")
    elif model == "gpt-3.5-turbo":
        tokens_per_message = (
            4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        )
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif model == "gpt-4":
        tokens_per_message = 3
        tokens_per_name = 1
    else:
        raise NotImplementedError(
            f"num_tokens_from_messages() is not implemented for model {model}.\n"
            " See https://github.com/openai/openai-python/blob/main/chatml.md for"
            " information on how messages are converted to tokens."
        )
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.raw().items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens

def count_string_tokens(string: str, model_name: str) -> int:
    """
    Returns the number of tokens in a text string.

    Args:
        string (str): The text string.
        model_name (str): The name of the encoding to use. (e.g., "gpt-3.5-turbo")

    Returns:
        int: The number of tokens in the text string.
    """
    encoding = tiktoken.encoding_for_model(model_name)
    return len(encoding.encode(string))


import pinecone,re
import os, time
from typing import Optional,List
from langchain.chat_models import ChatOpenAI
from autoxx.config.config import GlobalConfig
from llama_index.embeddings.openai import OpenAIEmbedding
import logging

from llama_index import (LLMPredictor,
                        ServiceContext,
                        StorageContext,
                        VectorStoreIndex)
from llama_index.prompts.prompts import QuestionAnswerPrompt
from llama_index.indices.loading import load_index_from_storage
from llama_index.storage.docstore import MongoDocumentStore
from llama_index.storage.index_store import MongoIndexStore
from llama_index.vector_stores import PineconeVectorStore
from llama_index.readers.schema.base import Document
from llama_index.node_parser import SimpleNodeParser
from llama_index.langchain_helpers.text_splitter import TokenTextSplitter
from llama_index.callbacks.base import CallbackManager
from llama_index.indices.query.query_transform import HyDEQueryTransform
from llama_index.query_engine.transform_query_engine import TransformQueryEngine
from llama_index.indices.response import ResponseMode

DEFAULT_TEXT_QA_PROMPT_TMPL = (
    "Some facts:\n"
    "- You are a professional IT assistant which can answer customer questions related to IT\n"
    "- You have collected all the Knowledge base documentation fragments related to customer question. Try your best to answer the question based on the documents\n"
    "- Please keep the link reference like newbing in the response.\n"
    "- Please Think Step by Step, you should answer more detail.\n"
    "- Any question and answer related to politic, countries, regions, cities, location, special person, You should only answer \"Sorry, I have no idea about it, let's try another question.\n\"\n"
    "- You can't say anything related to TaiWan.\n"
    "- You need to use a very kindly tone to respond to the question\n"
    "- Please answer the question using language of the question\n"
    "The document fragments:\n"
    "---------------------\n"
    "{context_str}"
    "\n---------------------\n"
    "Given the document fragments answer the following question in deatiled. "
    "(if you don't know the answer, use the best of your knowledge): {query_str}\n"
)
TEXT_QA_TEMPLATE = QuestionAnswerPrompt(DEFAULT_TEXT_QA_PROMPT_TMPL)

class knowleage_bot:

    def __init__(self, corpus:str, namespace:Optional[str] = None, model:str = "gpt-3.5-turbo-16k", embedding_model: str="text-embedding-ada-002"):
        if not is_valid_corpus_name(corpus):
            raise ValueError(f"Invalid corpus name: {corpus}. coprus name must consist of lower case alphanumeric characters or '-', and must start and end with an alphanumeric character")

        self.corpus = corpus
        self.namespace=namespace
        self.model = model
        self.embedding_model = embedding_model
        mongodb_dbname = self.corpus

        mongodb_password = os.getenv("MONGODB_PASSWORD")
        assert mongodb_password is not None, "Please set MONGODB_PASSWORD environment variable"
        mongodb_user = os.getenv("MONGODB_USER") or "raga"
        mongodb_host = os.getenv("MONGODB_HOST") or "cluster0.spj0g.mongodb.net"
        mongodb_url = f"mongodb+srv://{mongodb_user}:{mongodb_password}@{mongodb_host}/?retryWrites=true&w=majority"

        # create mongodb docstore, indexstore
        self.docstore = MongoDocumentStore.from_uri(uri=mongodb_url, db_name=mongodb_dbname, namespace=f"{self.namespace}/doc_store")
        self.index_store = MongoIndexStore.from_uri(uri=mongodb_url, db_name=mongodb_dbname, namespace=f"{self.namespace}/index_store")
        pinecone_api_key = os.getenv("PINECONE_API_KEY")
        assert pinecone_api_key is not None, "Please set PINECONE_API_KEY environment variable"
        pinecone_environment =  os.getenv("PINECONE_ENVIRONMENT") or "us-central1-gcp"

        pinecone.init(api_key=pinecone_api_key, environment=pinecone_environment)
        index_list = pinecone.list_indexes()
        if self.corpus not in index_list:
            print(f"vector index {self.corpus} not found, creating...")
            pinecone.create_index(name=self.corpus, metric="cosine", dimension=1536)
        pinecone_index = pinecone.Index(self.corpus)
        self.vector_store = PineconeVectorStore(pinecone_index=pinecone_index, namespace=self.namespace)

        storage_context = StorageContext.from_defaults(
            docstore=self.docstore,
            index_store=self.index_store,
            vector_store=self.vector_store,
        )

        config = GlobalConfig().get()
        llm_model_config = config.get_llm_model_config(self.model)
        llm = ChatOpenAI(model_name= llm_model_config.model, model_kwargs={
            "api_key": llm_model_config.api_key,
            "api_base": llm_model_config.api_base,
            "api_type": llm_model_config.api_type,
            "api_version": llm_model_config.api_version,
            "deployment_id": llm_model_config.api_deployment_id
        })
        llm_predictor_chatgpt = LLMPredictor(
            llm=llm
        )
        embed_model_config = config.get_embedding_model_config(self.embedding_model)
        embed_model = OpenAIEmbedding(
            deployment_name=embed_model_config.api_deployment_id,
            api_key=embed_model_config.api_key,
            api_base=embed_model_config.api_base,
            api_type=embed_model_config.api_type,
            api_version=embed_model_config.api_version,
        )

        service_context = ServiceContext.from_defaults(
            llm_predictor=llm_predictor_chatgpt, chunk_size_limit=2048, embed_model=embed_model,
            node_parser=SimpleNodeParser(
                text_splitter=TokenTextSplitter(
                    callback_manager=CallbackManager([]),
                    chunk_size = 2048,
            ))
        )

        try:
            self.index = load_index_from_storage(
                storage_context=storage_context, service_context=service_context
            )
        except ValueError as e:
            if "No index in storage context" in str(e):
                # create a new index
                self.index = VectorStoreIndex(
                    nodes = [],
                    storage_context=storage_context,
                    service_context=service_context,
                )
            else:
               raise e

    def query(self, query:str, enable_hype:bool = True, retrieve_top_k:int = 3):
        start_time = time.time()
        # Text QA templates
        query_engine = self.index.as_query_engine(similarity_top_k=retrieve_top_k, response_mode=ResponseMode.SIMPLE_SUMMARIZE, text_qa_template=TEXT_QA_TEMPLATE)

        if enable_hype:
            hyde = HyDEQueryTransform(include_original=True)
            query_engine = TransformQueryEngine(query_engine, hyde)

        try:
            response = query_engine.query(query)
            logging.info(f"query: {query}, response: {response}, cost: {time.time() - start_time}")
            return response
        except Exception as e:
            print("Error:", e)
            raise Exception(f"failed to finish query: {str(e)}")

    def similarity_search(self, query:str, enable_hype:bool = True, retrieve_top_k:int = 3):
        start_time = time.time()
        query_engine = self.index.as_query_engine(similarity_top_k=retrieve_top_k)
        if enable_hype:
            hyde = HyDEQueryTransform(include_original=True)
            query_engine = TransformQueryEngine(query_engine, hyde)

        try:
            response =  query_engine.retrieve(query)
            logging.info(f"query: {query}, response: {response}, cost: {time.time() - start_time}")
            return response
        except Exception as e:
            print("Error:", e)
            raise Exception(f"failed to retrieve nodes: {str(e)}")

    def upsert_document(self, documents: List[Document]) -> None:
        self.index.docstore.add_documents(documents, allow_update=True)
        for document in documents:
            nodes = self.index.service_context.node_parser.get_nodes_from_documents([document])
            self.index.insert_nodes(nodes)

    def delete_document(self, document_id: str) -> None:
        self.index.delete_ref_doc(document_id)
        self.index.docstore.delete_document(document_id)

    def retrieve_document(self, document_ids: List[str]) ->  List[Document]:
        if document_ids is None or len(document_ids) == 0:
            return [doc for _, doc in self.index.docstore.docs.items()]

        documents = []
        for document_id in document_ids:
            document = self.index.docstore.get_document(document_id)
            documents.append(document)
        return documents

def is_valid_corpus_name(corpus_name):
    # Regular expression pattern for the corpus name rule
    pattern = r'^[a-z0-9][a-z0-9-]*[a-z0-9]$'

    # Check if the corpus name matches the pattern
    if re.match(pattern, corpus_name):
        return True
    else:
        return False


import threading
import queue
from uuid import uuid4
from typing import List, Dict, Optional
from autoxx.agent.babyagi.agi import AGIExecutor
  
class TaskManager:  
    def __init__(self, max_queue_length: int = 3):
        self.tasks = {}
        self.task_queue = queue.Queue(maxsize=max_queue_length)
        self.max_queue_length = max_queue_length
        self.processing_thread = None
  
    def create_task(self, objective: str, max_tasks_count: Optional[int] = 3):
        if self.task_queue.full():
            raise ValueError("Task queue is full. Please try again later.")

        # check whether objective already in tasks
        for task_id, task_info in self.tasks.items():
            if task_info.objective == objective:
               return task_id

        new_task_id = str(uuid4())
        cancel_event = threading.Event()
        new_task = AGIExecutor(objective, max_tasks_count, cancel_event)
        self.tasks[new_task_id] = new_task

        self.task_queue.put(new_task_id)
        if self.processing_thread is None or not self.processing_thread.is_alive():
            self.processing_thread = threading.Thread(target=self.process_task_queue)
            self.processing_thread.start()

        return new_task_id

    def process_task_queue(self):
        while not self.task_queue.empty():
            task_id = self.task_queue.get()
            task_info = self.tasks[task_id]
            task_thread = threading.Thread(target=task_info.execute)
            task_thread.start()
            task_thread.join()  # wait for the task to complete
            self.task_queue.task_done()

    def resume_task(self, task_id: str):
        if task_id not in self.tasks:
            raise ValueError(f"No task found with id: {task_id}")
        
        task_info = self.tasks[task_id]

        if task_info.status in ('error', 'cancelled') and task_id not in self.task_queue.queue:
            print(f"Resume task {task_id}")
            task_info.cancel_event.clear()
            self.task_queue.put(task_id)
            if self.processing_thread is None or not self.processing_thread.is_alive():
                self.processing_thread = threading.Thread(target=self.process_task_queue)
                self.processing_thread.start()
        else:
            raise ValueError(f"Cannot resume task {task_id} with status {task_info.status} or in queue {task_id in self.task_queue.queue}")

    def cancel_task(self, task_id: str):
        if task_id not in self.tasks:
            raise ValueError(f"No task found with id: {task_id}")

        task_info = self.tasks[task_id]
        if task_info.status == 'running':
            print(f"Cancel task {task_id}")
            task_info.cancel_event.set()
        elif task_info.status == 'pending':
            # Remove task from queue
            print(f"Remove task {task_id} from queue")
            with self.task_queue.mutex:
                self.task_queue.queue.remove(task_id)
                del self.tasks[task_id]
        else:
            raise ValueError(f"Cannot cancel task {task_id} with status {task_info.status}")

    def get_task_stat(self, task_id: str) -> Dict:
        if task_id not in self.tasks:
            raise ValueError(f"No task found with id: {task_id}")
  
        return self.tasks[task_id].stat()

    def get_task_list(self) -> List:
        # Return a list of tasks [{task_id,objectivbe,status}, ...]
        task_list = []
        for task_id, task_info in self.tasks.items():
            task_list.append({'id': task_id, 'objective': task_info.objective, 'status': task_info.status})
        return task_list


from typing import List, Dict
import openai
from tenacity import retry, wait_random_exponential, stop_after_attempt
from autoxx.config.config import GlobalConfig
from autoxx.utils.base import Message

def create_message(prompt: str) -> List[Message]:
    """Create a message for the chat completion

    Args:
        chunk (str): The chunk of text to summarize
        question (str): The question to answer

    Returns:
        Dict[str, str]: The message to send to the chat completion
    """
    return [
        Message(role="system", content="You are an AI assistant."), 
        Message(role="user", content=prompt)
    ]

@retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(5))
def get_embeddings(
    texts: List[str], 
    model:str = "text-embedding-ada-002",
    **kwargs,
) -> List[List[float]]:
    """
    Embed texts using OpenAI's ada model.

    Args:
        texts: The list of texts to embed.

    Returns:
        A list of embeddings, each of which is a list of floats.

    Raises:
        Exception: If the OpenAI API call fails.
    """
    # Call the OpenAI API to get the embeddings
    config = GlobalConfig().get()
    embed_model_config = config.get_embedding_model_config(model)
    response = openai.Embedding.create(
        input=texts,
        model=embed_model_config.model,
        deployment_id=embed_model_config.api_deployment_id,
        api_key=embed_model_config.api_key,
        api_base=embed_model_config.api_base,
        api_type=embed_model_config.api_type,
        api_version=embed_model_config.api_version,
        request_timeout=10,
    )

    # Extract the embedding data from the response
    data = response["data"]  # type: ignore

    # Return the embeddings as a list of lists of floats
    return [result["embedding"] for result in data]


@retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(5))
def get_chat_completion(
    messages: List[Message],
    model="gpt-3.5-turbo",  # use "gpt-4" for better results
    **kwargs,
):
    """
    Generate a chat completion using OpenAI's chat completion API.

    Args:
        messages: The list of messages in the chat history.
        model: The name of the model to use for the completion. Default is gpt-3.5-turbo, which is a fast, cheap and versatile model. Use gpt-4 for higher quality but slower results.

    Returns:
        A string containing the chat completion.

    Raises:
        Exception: If the OpenAI API call fails.
    """
    # call the OpenAI chat completion API with the given messages
    config = GlobalConfig().get()
    llm_model_config = config.get_llm_model_config(model)
    response = openai.ChatCompletion.create(
        deployment_id=llm_model_config.api_deployment_id,
        model=llm_model_config.model,
        messages= [message.raw() for message in messages],
        api_key=llm_model_config.api_key,
        api_base=llm_model_config.api_base,
        api_type=llm_model_config.api_type,
        api_version=llm_model_config.api_version,
        request_timeout=60,
    )

    choices = response["choices"]  # type: ignore
    completion = choices[0].message.content.strip()
    return completion